from pathlib import Path
from typing import List

from loguru import logger
import orjson
import polars as pl
from tqdm import tqdm

from used_addr_check.index_types import IndexEntry


def generate_index(
    haystack_file_path: Path, index_chunk_size: int = 1000
) -> List[IndexEntry]:
    """
    Generates an index for a large sorted text file, storing every
    `index_chunk_size` lines.

    Args:
    - haystack_file_path (Path): Path to the file to be indexed.
    - index_chunk_size: The number of lines to store in each index entry.

    Returns:
    - List[IndexEntry]: A list of tuples containing
            line text, byte offset, and line number.
    """
    index: List[IndexEntry] = []
    haystack_file_size = haystack_file_path.stat().st_size
    with (
        tqdm(
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
            total=haystack_file_size,
            desc="Indexing haystack file",
        ) as progress_bar,
        open(haystack_file_path, "rb") as file,
    ):
        offset = 0
        for line_number, line in enumerate(file):
            line_len = len(line)
            if line_number % index_chunk_size == 0:
                index.append(
                    IndexEntry(
                        line_value=line.strip().decode("utf-8"),
                        byte_offset=offset,
                        line_number=line_number,
                    )
                )

                progress_bar.n = offset
                progress_bar.refresh()

            offset += line_len
    return index


def store_index_json(
    index: List[IndexEntry], index_json_file_path: Path
) -> None:
    """
    Stores the index in a file for later use.

    Args:
    - index (List[IndexEntry]): The index to store.
    - index_file_path (Path): The path to store the index.
    """

    with open(index_json_file_path, "wb") as file:
        file.write(orjson.dumps(index))


def load_index_json(index_json_file_path: Path) -> List[IndexEntry]:
    """
    Loads an index from a JSON file.

    Args:
    - index_json_file_path (Path): The path to the JSON file.

    Returns:
    - List[IndexEntry]: The loaded index.
    """
    with open(index_json_file_path, "rb") as file:
        raw_read: List[dict] = orjson.loads(file.read())
    return [IndexEntry(**val) for val in raw_read]


def store_index_parquet(
    index: List[IndexEntry], index_parquet_file_path: Path
) -> None:
    """
    Stores the index in a parquet file for later use.

    Args:
    - index (List[IndexEntry]): The index to store.
    - index_file_path (Path): The path to store the index.
    """

    df = pl.DataFrame(index)
    df.write_parquet(index_parquet_file_path)


def load_index_parquet(index_parquet_file_path: Path) -> List[IndexEntry]:
    """
    Loads an index from a parquet file.

    Args:
    - index_parquet_file_path (Path): The path to the parquet file.

    Returns:
    - List[IndexEntry]: The loaded index.
    """
    df = pl.read_parquet(index_parquet_file_path)
    return [IndexEntry(**val) for val in df.to_dicts()]


def load_or_generate_index(
    haystack_file_path: Path,
    index_chunk_size: int = 1000,
    force_recreate: bool = False,
) -> List[IndexEntry]:
    """Attempts to load an index from a file, or generates one if it doesn't,
    or if `force_recreate` is enabled.

    Tries to load the index from a Parquet file first, then from a JSON file.

    If a file already exists, the `index_chunk_size` is ignored.
    """
    index_json_file_path = haystack_file_path.with_suffix(".index.json")
    index_parquet_file_path = haystack_file_path.with_suffix(".index.parquet")

    if force_recreate or (
        not index_json_file_path.exists()
        and not index_parquet_file_path.exists()
    ):
        logger.info(f"Creating index for file: {haystack_file_path.name}")
        index = generate_index(
            haystack_file_path, index_chunk_size=index_chunk_size
        )
        logger.info(f"Index created with {len(index):,} entries")

        # store to main type (parquet)
        store_index_parquet(index, index_parquet_file_path)
        logger.info(
            f"Index stored in {index_parquet_file_path.name}, "
            f"size: {index_parquet_file_path.stat().st_size:,} bytes"
        )
        return index

    if index_parquet_file_path.exists():
        logger.info("Loading index from Parquet file")
        index = load_index_parquet(index_parquet_file_path)
        logger.info(f"Index loaded with {len(index):,} entries")
        return index

    if index_json_file_path.exists():
        logger.info("Loading index from JSON file")
        index = load_index_json(index_json_file_path)
        logger.info(f"Index loaded with {len(index):,} entries")
        return index

    raise Exception("This should be unreachable.")



# from used_addr_check.download_list import download_list, BITCOIN_LIST_URL
from used_addr_check.index_create import load_or_generate_index
from used_addr_check.index_search import search_multiple_in_file
from used_addr_check.scan_file import scan_file_for_used_addresses

import argparse
from pathlib import Path


def main_cli():
    parser = argparse.ArgumentParser(
        description="CLI for file processing and searching"
    )
    subparsers = parser.add_subparsers(dest="command")

    # # Subparser for the 'download' command
    # download_parser = subparsers.add_parser(
    #     "download", help="Download the file"
    # )
    # download_parser.add_argument(
    #     "-o",
    #     "--output",
    #     dest="output_path",
    #     required=True,
    #     help="Output file path (should end in .gz)",
    # )
    # download_parser.add_argument(
    #     "-u",
    #     "--url",
    #     dest="url",
    #     default=BITCOIN_LIST_URL,
    #     help="URL to download the file from",
    # )

    # Subparser for the 'index' command
    index_parser = subparsers.add_parser(
        "index",
        help="Index a haystack 'used addresses' file, save it to orig_name.txt.index.json",  # noqa
    )
    index_parser.add_argument(
        "-f",
        "--haystack",
        dest="haystack_file_path",
        required=True,
        help="Haystack address list file path (.txt)",
    )

    # Subparser for the 'search' command
    search_parser = subparsers.add_parser("search", help="Search a file")
    search_parser.add_argument(
        "-f",
        "--haystack",
        dest="haystack_file_path",
        required=True,
        help="Haystack address list file path (.txt)",
    )
    search_parser.add_argument(
        "-n",
        "--needle",
        dest="needles",
        required=True,
        action="append",
        help="Search query(s) to find in the file",
    )

    # Subparser for the 'scan_file' command
    scan_file_parser = subparsers.add_parser(
        "scan_file",
        help="Scan a file for bitcoin addresses, and see which ones have been used.",  # noqa
    )
    scan_file_parser.add_argument(
        "-f",
        "--haystack",
        dest="haystack_file_path",
        required=True,
        help="Haystack address list file path (.txt)",
    )
    scan_file_parser.add_argument(
        "-n",
        "--needle",
        dest="needle_haystack_file_path",
        required=True,
        help="Needle file path, with list of addresses. Addresses will be "
        "extracted from this file",
    )

    args = parser.parse_args()

    if args.command == "index":
        load_or_generate_index(
            haystack_file_path=Path(args.haystack_file_path),
            force_recreate=True,
        )
    elif args.command == "search":
        search_multiple_in_file(Path(args.haystack_file_path), args.needles)
    # elif args.command == "download":
    #     download_list(Path(args.output_path))
    elif args.command == "scan_file":
        scan_file_for_used_addresses(
            Path(args.haystack_file_path),
            Path(args.needle_haystack_file_path),
        )
    else:
        parser.print_help()


if __name__ == "__main__":
    main_cli()



from pathlib import Path
from typing import List
import re

from loguru import logger

from used_addr_check.index_search import search_multiple_in_file


def extract_addresses_from_file(text_file_path: Path) -> List[str]:
    """
    Extracts bitcoin addresses from a file.

    Args:
    - text_file_path (Path): The path to the file to extract addresses from.

    Returns:
    - List[str]: A list of bitcoin addresses found in the file.
    """
    assert isinstance(text_file_path, Path)

    # TODO: implement chunk scanning for large files
    # TODO: add support for weird address formats (e.g., the ones at the
    # tail of the address list file)

    with text_file_path.open("r") as file:
        return re.findall(r"[13][a-km-zA-HJ-NP-Z1-9]{25,34}", file.read())


def scan_file_for_used_addresses(
    haystack_file_path: Path, needle_file_path: Path
):
    """
    Scans a file for bitcoin addresses, and see which one have been used.

    Args:
    - haystack_file_path (Path): The path to the file to scan.
    - needle_file_path (Path): The path to the file with the list of addresses
        to search for in the haystack file.
    """
    assert isinstance(haystack_file_path, Path)
    assert isinstance(needle_file_path, Path)

    needle_addresses = extract_addresses_from_file(needle_file_path)
    logger.info(
        f"Extracted {len(needle_addresses):,} addresses from the needle file"
    )
    matched_addresses = search_multiple_in_file(
        haystack_file_path, needle_addresses
    )
    logger.info(f"Found {len(matched_addresses):,} used addresses in the file")



import bisect
from pathlib import Path
from typing import List

from loguru import logger
from tqdm import tqdm

from used_addr_check.index_create import load_or_generate_index
from used_addr_check.index_types import IndexEntry


def _binary_search_index(index: List[IndexEntry], needle: str) -> int:
    """
    Performs a binary search on the index to find the closest position for the
    needle string.

    Args:
    - index (List[IndexEntry]): The index list, as returned by `create_index()`
    - needle (str): The string to find in the index.

    Returns:
    - int: The index position (index into `index` list), to the left of where
        the needle could/would be located.
    """
    index_values = [entry.line_value for entry in index]
    idx = bisect.bisect_left(index_values, needle)

    # not sure what this condition is about
    if idx == len(index):
        return idx - 1

    if index[idx].line_value > needle:
        # common occurrence
        return idx - 1
    elif index[idx].line_value == needle:
        # rare occurrence - exact match on one of the index keys (e.g., 0)
        return idx
    elif index[idx].line_value < needle:
        # not sure if this can ever happen
        raise Exception(
            "bisect.bisect_left failed to find the correct position"
        )
    else:
        raise Exception("Unexpected condition")


def search_in_file_with_index(
    haystack_file_path: Path, needle: str, index: List[IndexEntry]
) -> bool:
    """
    Searches for a needle string in the file using a pre-built index to narrow
    down the search area.

    Args:
    - haystack_file_path (Path): The path to the file to search.
    - needle (str): The string to search for in the file.
    - index (List[Tuple[str, int, int]]): The index as built by `create_index`.

    Returns: True if the `needle` string is found, False otherwise.
    """
    assert isinstance(haystack_file_path, Path)
    assert isinstance(needle, str)
    assert isinstance(index, list)

    position = _binary_search_index(index, needle)
    # logger.debug(f"Search {needle}: Binary search position: {position}")
    # logger.debug(f"Search {needle}: {index[position]}")
    if position == len(index):
        # logger.debug(
        #     "Binary search position equals index length, returning False"
        # )
        return False

    # Find the bounds to search within the file
    start_offset = index[position].byte_offset
    end_offset = None
    if position + 1 < len(index):
        end_offset = index[position + 1].byte_offset

    with open(haystack_file_path, "r", encoding="ascii") as file:
        file.seek(start_offset)
        while True:
            if end_offset and file.tell() >= end_offset:
                break
            line = file.readline()
            if not line:
                break
            if line.strip() == needle:
                return True
    return False


def search_multiple_in_file(
    haystack_file_path: Path | str,
    needles: List[str] | str,
    index_chunk_size: int = 1000,
) -> List[str]:
    """
    Searches for multiple needle strings in the file by pre-building an index
    and then searching within the file.

    Args:
    - haystack_file_path (Path): The path to the file to search.
    - needles (List[str]): The list of strings to search for in the file.

    Returns: A list of the needles that were found in the file.
    """
    if isinstance(needles, str):
        needles = [needles]
    if isinstance(haystack_file_path, str):
        haystack_file_path = Path(haystack_file_path)

    index = load_or_generate_index(haystack_file_path, index_chunk_size)

    # do the search
    found_needles = []
    for needle in tqdm(needles, desc="Searching needles", unit="needle"):
        if search_in_file_with_index(haystack_file_path, needle, index=index):
            found_needles.append(needle)
    logger.info(
        f"Found {len(found_needles):,}/{len(needles):,} needles in the file"
    )
    logger.info(f"Needles found: {found_needles}")
    return found_needles



__VERSION__ = "0.0.1"
__AUTHOR__ = "RecRanger"

from .index_create import (  # noqa F401
    load_or_generate_index,
    generate_index,
    store_index_json,
    load_index_json,
    store_index_parquet,
    load_index_parquet,
)
from .index_search import (  # noqa F401
    search_in_file_with_index,
    search_multiple_in_file,  # <- main library function
)
from .index_types import IndexEntry  # noqa F401
from .cli import main_cli  # noqa F401



from pathlib import Path

import backoff
from loguru import logger
import requests

# TODO: This download feature isn't really production-ready.
# TODO: Make the downloader as fast as wget.
# TODO: Create an extract function.


BITCOIN_LIST_URL = "http://alladdresses.loyce.club/all_Bitcoin_addresses_ever_used_sorted.txt.gz"  # noqa


def _get_remote_file_size(url: str) -> int:
    with requests.head(url) as r:
        r.raise_for_status()
        return int(r.headers.get("content-length", 0))


@backoff.on_exception(
    backoff.constant,
    requests.exceptions.RequestException,
    max_tries=30,
    interval=5,
    on_backoff=lambda details: logger.debug(
        f"Transfer failed partway; resuming... (try {details['tries']}/30)"
    ),
)
def _download_file(url: str, dest: Path) -> Path:
    """Download a file from a URL to a local destination, resuming if possible.
    If the file already exists and is fully downloaded, do nothing.
    If the file is partially downloaded, resume the download.
    If the file does not exist, download it from scratch.
    If `dest` is a directory, the file will be saved in that directory.

    Returns the path to the downloaded file.
    """
    total_size = _get_remote_file_size(url)

    if dest.is_dir():
        destination = dest / Path(url).name
    else:
        destination = dest

    # Check how much of the file has already been downloaded
    if destination.exists():
        current_size = destination.stat().st_size
    else:
        current_size = 0

    # If file is already fully downloaded
    if current_size == total_size:
        logger.info("File already downloaded.")
        return destination
    elif current_size > total_size:
        logger.warning(
            f"File size mismatch. Current size: {current_size:,} bytes, "
            f"Total size: {total_size:,} bytes"
        )
        logger.warning("Deleting the file and re-downloading...")
        destination.unlink()
        current_size = 0
    else:
        # File is partially downloaded
        logger.info(
            f"Resuming download at offset={current_size:,}/{total_size:,}"
        )

    # Stream the file from the last byte we have already downloaded
    headers = {"Range": f"bytes={current_size}-"}
    with requests.get(url, headers=headers, stream=True) as r:
        r.raise_for_status()
        # Open the file in append mode, create if does not exist
        with open(destination, "ab") as f:
            for chunk in r.iter_content(chunk_size=8192):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)
                    f.flush()
    return destination


def download_list(destination: Path) -> None:
    logger.info(f"Downloading list from: {BITCOIN_LIST_URL}")
    _download_file(BITCOIN_LIST_URL, destination)
    logger.info(
        f"Downloaded list to: {destination}. "
        f"Size: {destination.stat().st_size:,} bytes."
    )
